__import__("sys").path[0:0] =  ["."]

from ..mocodo_error import MocodoError
from ..parse_mcd import Token, Visitor, Transformer, Lark_StandAlone, Discard

def parse_source(source):
    parser = Lark_StandAlone()
    try:
        return parser.parse(f"{source}\n")
    except Exception as error:
        pin = _("Parsing error:\n{e}\n").format(e=error.get_context(source))
        expected = set(error.expected)
        t = error.token.type
        v = repr(error.token.value)
        if error.token.value == "\n":
            t = "BREAK"
        try:
            previous = error.token_history[0].type
        except:
            previous = None
        if expected == {'PERCENT', 'CONSTRAINT_LPAREN', 'BREAK', 'INDENT', 'PHANTOMS', 'BOX_NAME', 'SLASH', 'PLUS', 'MINUS', 'NL'}:
            raise MocodoError(501, _('{pin}"{v}" is not a valid line beginning.').format(pin=pin, v=v)) # fmt: skip
        if expected == {'PHANTOMS', 'PERCENT', 'CONSTRAINT_LPAREN', 'PLUS', 'MINUS', 'NL', 'SLASH', 'BOX_NAME'}:
            raise MocodoError(501, _('{pin}"{v}" is not a valid line beginning.').format(pin=pin, v=v)) # fmt: skip
        if expected == {'INHERITANCE_ARROW', 'COLON', 'NL', 'MORETHAN', 'SP', 'COMMA'}:
            raise MocodoError(502, _('{pin}Malformed box name.').format(pin=pin, v=v)) # fmt: skip
        if expected == {'COLON', 'COMMA'}:
            raise MocodoError(503, _('{pin}A valid box name starting a line must be followed by a colon or a comma.').format(pin=pin)) # fmt: skip
        if t == "COMMA" and expected == {'SP', 'BOX_NAME'}:
            raise MocodoError(505, _('{pin}Illegal comma after inheritance.').format(pin=pin)) # fmt: skip
        if t == "NUMBER" and expected == {'BOX_NAME', 'CARD', 'LBRACKET', 'LEG_ARROW', 'MINUS', 'SLASH', 'SP', 'ID_MARK'}:
            raise MocodoError(506, _('{pin}Malformed cardinalities.').format(pin=pin)) # fmt: skip
        if expected == {'CARD'}:
            raise MocodoError(506, _('{pin}Malformed cardinalities.').format(pin=pin)) # fmt: skip
        if expected == {'SLASH', 'CARD', 'ID_MARK'}:
            raise MocodoError(506, _('{pin}Malformed cardinalities.').format(pin=pin)) # fmt: skip
        if expected == {'INHERITANCE_NAME', 'BACKSLASH'} or expected == {'BACKSLASH'}:
            raise MocodoError(507, _('{pin}An inheritance name must be "", "X", "T" or "XT".').format(pin=pin)) # fmt: skip
        if t == "COMMA" and expected == {'NL'}:
            raise MocodoError(508, _('{pin}Only two coords are allowed.').format(pin=pin)) # fmt: skip
        if t == "BREAK" and expected == {'BOX_NAME', 'CARD', 'LBRACKET', 'LEG_ARROW', 'MINUS', 'SLASH', 'SP', 'ID_MARK'}:
            raise MocodoError(509, _('{pin}An association leg cannot be empty.').format(pin=pin)) # fmt: skip
        if expected == {'BOX_NAME', 'NUMBER'}:
            raise MocodoError(510, _('{pin}Expected a number or a box name.').format(pin=pin)) # fmt: skip
        if expected == {'BOX_NAME'}:
            raise MocodoError(511, _('{pin}Only a box name is possible here.').format(pin=pin)) # fmt: skip
        if expected == {'RBRACKET'}:
            raise MocodoError(512, _('{pin}Unclosed square bracket.').format(pin=pin)) # fmt: skip
        if expected == {'CONSTRAINT_RPAREN'}:
            raise MocodoError(514, _('{pin}A constraint name cannot contain more than three characters.').format(pin=pin)) # fmt: skip
        if expected == {'SP', 'BOX_NAME', 'CONSTRAINT_LEG'}:
            raise MocodoError(515, _('{pin}Expected a box name or a constraint leg.').format(pin=pin)) # fmt: skip
        if expected == {'LBRACKET', 'SP', 'BOX_NAME', 'NL', 'COLON', 'CONSTRAINT_LEG'}:
            raise MocodoError(516, _('{pin}Illegal character after a constraint name.').format(pin=pin)) # fmt: skip
        if expected == {'INHERITANCE_ARROW', 'SP'} or expected == {'INHERITANCE_ARROW'}:
            raise MocodoError(517, _('{pin}A parent name must be followed by an inheritance arrow among "<=", "<-", "->", "=>".').format(pin=pin)) # fmt: skip
        if t == "INHERITANCE_ARROW" and expected == {'MORETHAN'}:
            raise MocodoError(518, _('{pin}Please change the old foreign key syntax ("->") by the new one (">").').format(pin=pin)) # fmt: skip
        if t == "SP" and expected == {'COMMA', 'COLON', 'NL'}:
            raise MocodoError(519, _('{pin}The constraint targets must be comma-separated.').format(pin=pin)) # fmt: skip
        if expected == {'HASHTAG', 'NL', 'ID_GROUPS', 'ID_MARK', 'ATTR', 'COMMA'}:
            raise MocodoError(500, _('{pin}An attribute label cannot start with "{v[1]!r}".').format(pin=pin, v=v)) # fmt: skip
        if expected == {'ID_MARK', 'NL', 'ATTR', 'COMMA'}:
            raise MocodoError(500, _('{pin}An attribute label cannot start with "{v[1]!r}".').format(pin=pin, v=v)) # fmt: skip
        if expected == {'LBRACKET', 'NL', 'COMMA'}:
            raise MocodoError(521, _('{pin}An attribute label cannot contain "{v}".').format(pin=pin, v=v)) # fmt: skip
        if t in ("COMMA", "NL", "BREAK") and expected == {'MORETHAN'}:
            raise MocodoError(522, _('{pin}An attribute starting with "#" must contain two ">".').format(pin=pin)) # fmt: skip
        if t in ("COMMA", "NL", "BREAK") and expected == {'ATTR'}:
            raise MocodoError(523, _('{pin}Expected an entity name.').format(pin=pin)) # fmt: skip
        if expected == {'COLON', 'COMMA', 'NL'}:
            raise MocodoError(524, _('{pin}A box name cannot contain "{v}".').format(pin=pin, v=v)) # fmt: skip
        if expected == {"COMMA"}:
            raise MocodoError(525, _('{pin}Expected a comma.').format(pin=pin, v=v)) # fmt: skip)
        if t == "BOX_NAME" and expected == {'COMMA', 'NL'}:
            raise MocodoError(526, _('{pin}Malformed number.').format(pin=pin, v=v)) # fmt: skip
        if previous in ("NUMBER", "BOX_NAME") and expected == {'COMMA', 'NL'}:
            raise MocodoError(527, _('{pin}More than two coordinates.').format(pin=pin, v=v)) # fmt: skip
        if expected == {'LBRACKET', 'NL', 'COMMA', 'MORETHAN'}:
            raise MocodoError(528, _('{pin}An attribute label cannot have more than one optionality marker.').format(pin=pin, v=v))
        raise MocodoError(504, _('{pin}Token "{t}" encountered. Expected tokens: {expected}.').format(pin=pin, expected=expected, t=t))
        
    
class Reconstructor(Visitor):

    def __init__(self):
        self.result = []
    
    def __default__(self, tree):
        for child in tree.children:
            if isinstance(child, Token):
                self.result.append((child.line, child.column, child.value))

def reconstruct_source(tree):
    visitor = Reconstructor()
    try:
        visitor.visit(tree)
    except Exception as visitor_error:
        raise visitor_error.__context__ if visitor_error.__context__ else visitor_error
    return "".join(s for (_, _, s) in sorted(visitor.result))

def transform_source(source, transformer):
    tree = parse_source(source)
    try:
        new_tree = transformer.transform(tree)
    except Exception as transformer_error:
        raise transformer_error.__context__ if transformer_error.__context__ else transformer_error
    return reconstruct_source(new_tree)

class ClauseExtractor(Transformer):

    discard = lambda self, tree: Discard
    COLON = discard
    COMMA = discard
    NL = discard
    SP = discard
    LBRACKET = discard
    RBRACKET = discard
    LPAREN = discard
    RPAREN = discard
    CONSTRAINT_LPAREN = discard
    CONSTRAINT_RPAREN = discard
    HASHTAG = discard
    MORETHAN = discard
    PERCENT = discard
    BACKSLASH = discard

    def _item(self, key, tree):
        return (key, tree[0].value)
    
    card = lambda self, tree: self._item("card", tree)
    entity_name_ref = lambda self, tree: self._item("entity", tree)
    box_name_ref = lambda self, tree: self._item("name", tree)
    leg_note = lambda self, tree: self._item("leg_note", tree)
    constraint_note = lambda self, tree: self._item("constraint_note", tree)
    constraint_leg = lambda self, tree: self._item("constraint_leg", tree)
    constraint_name = lambda self, tree: self._item("name", tree)
    card_prefix = lambda self, tree: self._item("card_prefix", tree)
    card_hidden = lambda self, tree: self._item("card_hidden", tree)
    leg_arrow = lambda self, tree: self._item("leg_arrow", tree)
    inheritance_arrow = lambda self, tree: self._item("inheritance_arrow", tree)
    entity_name_def = lambda self, tree: self._item("name", tree)
    inheritance_name = lambda self, tree: self._item("name", tree)
    assoc_name_def = lambda self, tree: self._item("name", tree)
    attr = lambda self, tree: self._item("attribute_label", tree)
    box_def_prefix = lambda self, tree: self._item("box_def_prefix", tree)
    id_mark = lambda self, tree: self._item("id_mark", tree)

    def start(self, tree):
        return tree

    def line(self, tree):
        d = {}
        for t in tree:
            d.update(t)
        return d
    
    def break_(self, tree):
        return {"type": "break"}

    def phantoms(self, tree):
        return {"type": "phantoms", "count": tree[0].value.count(":")}
    
    def comment(self, tree):
        return {"type": "comment", "text": f"%{tree[0].value.rstrip()}"}
    
    def entity_clause(self, tree):
        return tree + [("type", "entity")]

    def assoc_clause(self, tree):
        return tree + [("type", "association")]
    
    def constraint_clause(self, tree):
        return tree + [("type", "constraint")]
    
    def inheritance_clause(self, tree):
        # Tweak the tree to make it look like an association
        d = dict(item for item in tree if isinstance(item, tuple))
        d["name"] = d.get("name", "")
        d["type"] = "inheritance"
        d["legs"] = [{"entity": d["entity"], "rank": -1}] + d["legs"]
        del d["entity"]
        return list(d.items())

    def typed_attr(self, tree):
        return ("attr", dict(tree))
    
    def assoc_leg(self, tree):
        return ("leg", dict(tree))
    
    def inheritance_parent(self, tree):
        return ("entity", tree[0][1])
    
    def inheritance_child(self, tree):
        return ("leg", dict(tree))
    
    def foreign_reference(self, tree):
        return ("foreign_reference", dict(tree))

    def box_name(self, tree):
        return tree[0]
    
    def seq(self, tree):
        items = []
        for (rank, (_, d)) in enumerate(tree):
            d["rank"] = rank
            items.append(d)
        return (f"{tree[0][0]}s", items)

    def assoc_attr(self, tree):
        if tree[0] == "_":
            tree[-1][1]["id_mark"] = "_"
        return ("attr", tree[-1][1])

    def assoc_leg(self, tree):
        return ("leg", dict(tree))
    
    def constraint_target(self, tree):
        return ("constraint_target", dict(tree))
    
    def foreign_reference(self, tree):
        return ("foreign_reference", dict(tree))

    def entity_or_table_attr(self, tree):
        d = dict(tree)
        if "foreign_reference" in d:
            d.update(d.pop("foreign_reference"))
        if "attr" in d:
            d.update(d.pop("attr"))
        return ("attr", d)
    
    def this_table_attr(self, tree):
        return ("attribute_label", tree[0][1])
    
    def that_table(self, tree):
        return ("that_table", tree[0][1])
    
    def that_table_attr(self, tree):
        return ("that_table_attribute_label", tree[0][1])
    
    def constraint_coords(self, tree):
        return ("constraint_coords", [self._constraint_coord(x) for x in tree])
    
    def _constraint_coord(self, x):
        try:
            x = float(x)
            return int(x) if x == int(x) else x
        except (TypeError, ValueError):
            return x[1]

    def indent(self, tree):
        return {"type": "break", "indent": tree[0].value} # Add a default type for dangling indents
    
    def datatype(self, tree):
        return ("datatype", tree[0].value if tree else "")
    
    def id_groups(self, tree):
        return ("id_groups", "".join(sorted(set(tree[0].value))))
    
def extract_clauses(source):
    tree = parse_source(source)
    extractor = ClauseExtractor()
    result = extractor.transform(tree)
    for line in result:
        if "type" not in line:
            line["type"] = "break"
    # Uniformize indentations. First, find the set of all distinct indentations.
    # Include the empty string if absent. Then, sort them by length. Finally,
    # replace each indentation by a number of spaces proportional to its index
    # in the sorted list.
    indents = set(line.get("indent", "") for line in result)
    indents.add("")
    indents = sorted(indents, key=len)
    indents = {indent: "  " * i for (i, indent) in enumerate(indents)}
    for (line, raw_line) in zip(result, source.splitlines()):
        line["indent"] = indents[line.get("indent", "")]
        line["source"] = line["indent"] + raw_line.lstrip()
    return result


def first_child(tree, name, i=0):
    result = next(tree.find_data(name), None)
    if result is None:
        return ""
    return result.children[i]

def is_identifier(i, id_groups, id_mark):
    # Entity (foo, bar)
    if i == 0:
        if "0" not in id_groups:
            return True # foo / 1_foo
    elif "0" in id_groups:
        return True # 0_bar / 01_bar
    elif id_mark == "_" and id_groups == "":
        return True # _bar
    return False # 0_foo / 01_foo / _foo / bar / 1_bar

if __name__ == "__main__":
    # python -m mocodo.tools.parser_tools
    source = "EMPLOYER, 01 PARTICIPANT, 0N ENTREPRISE:\nCLIENT: num client, nom"
    tree = parse_source(source)
    print(tree.pretty())
    print(reconstruct_source(tree))
    print(extract_clauses(source))



import sys

if sys.version_info < (3, 6):
    print(f"Mocodo requires Python 3.6 or later to run.\nThis version is {sys.version}.")
    sys.exit()

import importlib
from pathlib import Path
import json

from .argument_parser import parsed_arguments
from .common import Common, Printer
from .convert.read_template import read_template
from .convert.relations import Relations
from .font_metrics import font_metrics_factory
from .guess_title import may_update_params_with_guessed_title
from .mcd import Mcd
from .mcd_to_svg import main as dump_mcd_to_svg
from .mocodo_error import MocodoError, subopt_error
from .rewrite import op_tk, guess_entities
from .tools.graphviz_tools import minify_graphviz
from .tools.string_tools import urlsafe_encoding
from .tools.various import invert_dict


SHOW_ARGS = invert_dict({
    "mcd": ["mcd"],
    "rw": ["rw", "source", "text", "code", "mocodo"],
    "cv": ["cv", "mld", "ddl", "sql"],
})

class ResponseLogger:

    def __init__(self, params):
        if not params["is_magic"]:
            self.may_log = self.log_nothing
            return
        self.response = {
            "mld": params["mld"],
            "args_to_delete": params["args_to_delete"],
            "opt_to_restore": params["opt_to_restore"],
            "redirect_output": params["redirect_output"],
            "select": params["select"],
        }
        self.may_log = self.log_for_magic
        self.path = Path(f"{params['output_name']}_response_for_magic_command.json")
        self.dump()
    
    def log_nothing(self, *args, **kwargs):
        pass

    def log_for_magic(self, key, value):
        self.response[key] = value
        self.dump()
    
    def dump(self):
        self.path.write_text(json.dumps(self.response, ensure_ascii=False), encoding="utf8")


def flip(source, subargs):
    for subsubopt in "".join(subargs.keys()):
        mcd = Mcd(source)
        if subsubopt == "v":
            source = mcd.get_vertically_flipped_clauses()
        elif subsubopt == "h":
            source = mcd.get_horizontally_flipped_clauses()
        elif subsubopt == "d":
            source = mcd.get_diagonally_flipped_clauses()
        else:
            subopt_error("flip", subsubopt)
    return source


class Runner:

    def __init__(self, args, printer):
        self.printer = printer
        try:
            self.params = parsed_arguments(args)
            self.parsing_error = None
            self.common = Common(self.params)
            self.get_font_metrics = font_metrics_factory(self.params)
        except MocodoError as e:
            # Raising a parsing error must be delayed, otherwise it will not be displayed as a MocodoError
            self.parsing_error = e
    
    def __call__(self):
        if self.parsing_error:
            raise self.parsing_error
        
        source = self.common.load_input_file()

        if self.params["restore"]:
            shutil = importlib.import_module("shutil")
            path = Path(self.params["script_directory"], "resources", "pristine_sandbox.mcd")
            shutil.copyfile(path, "sandbox.mcd")
            Path("params.json").write_text("{}\n", encoding="utf8")
            return
        
        if self.params["print_params"]:
            for added_key in self.params["keys_to_hide"][:]:
                self.params.pop(added_key, None)
            self.params["output_dir"] = str(Path(self.params["output_dir"]).resolve().relative_to(Path.cwd()))
            text = json.dumps(self.params, ensure_ascii=False, indent=2, sort_keys=True)
            text = text.replace("\n    ", " ")
            text = text.replace("\n  ]", " ]")
            self.printer.write(text.strip())
            return

        self.add_gutter_params(self.params)

        if self.params["mld"] or ("transform" in self.params and self.params["transform"] == []):
            # In case there is an option `--mld` or an option `--transform` without arguments,
            # inject manually the equivalent --convert sub-option
            self.params["mld"] = True
            self.params["convert"].insert(0, ("markdown", {}))

        if "select" in self.params: # the user wants to override the default display policy under Jupyter
            if "all" in self.params["select"]:
                self.params["select"] = ["mcd", "rw", "cv"] if self.params["rewrite"] else ["mcd", "cv"]
            normalized_user_choices = []
            for k in self.params["select"]:
                if k.lower() not in SHOW_ARGS:
                    raise MocodoError(28, _('Unknown argument "{k}" for option --select.').format(k=k)) # fmt: skip
                normalized_user_choices.append(SHOW_ARGS[k.lower()])
            self.params["select"] = normalized_user_choices
        else:
            if self.params["rewrite"] and self.params["convert"]:
                self.params["select"] = ["mcd", "cv"]
            elif self.params["rewrite"]:
                self.params["select"] = ["mcd", "rw"]
            elif self.params["convert"] and self.params["mld"]:
                self.params["select"] = ["mcd", "cv"]
            elif self.params["convert"]:
                self.params["select"] = ["cv"]
            else:
                self.params["select"] = ["mcd"]

        response = ResponseLogger(self.params)

        if self.params["rewrite"]:
            for (subopt, subargs) in self.params["rewrite"]:
                if subopt == "echo":
                    pass
                elif subopt == "flip":
                    source = flip(source, subargs)
                elif subopt == "create" and "entities" in subargs:
                    source = guess_entities.run(source, subargs["entities"])
                elif subopt == "delete" and "dfs" in subargs:
                    module = importlib.import_module(f".rewrite._delete_dfs", package="mocodo")
                    source = module.run(source, self.params)
                elif subopt in self.params["op_tk_rewritings"]: # ex.: create, delete, ascii, etc.
                    source = op_tk.run(source, op_name=subopt, subargs=subargs, params=self.params).rstrip()
                else: # An unspecified rewrite operation, dynamically loaded
                    try:
                        module = importlib.import_module(f".rewrite._{subopt}", package="mocodo")
                    except ModuleNotFoundError:
                        raise subopt_error("rewrite", subopt)
                    source = module.run(source, subopt=subopt, subargs=subargs, params=self.params).rstrip()
                response.may_log("rewritten_source", source)
            if not self.params["is_magic"]:
                self.printer.write(self.common.update_source(source))
        
        may_update_params_with_guessed_title(source, self.params)

        converted_file_paths = [] # list of files to be displayed in the notebook
        raw_relations = None
        sql_relations = None
        if self.params["convert"]:
            response.may_log("has_explicit_conversion", True)
            results = []
            for (subopt, subargs) in self.params["convert"]:

                # Try to interpret subopt as the stem or path of a relation template
                template = None
                official_template_dir = Path(self.params["script_directory"], "resources", "relation_templates")
                if subopt == "relation":
                    stem_or_path = next(iter(subargs.keys()), "") # ignore all sub-arguments after the first one
                    template = read_template(stem_or_path, official_template_dir)
                elif Path(official_template_dir, f"{subopt}.yaml").is_file():
                    template_suffix = next(iter(subargs.keys()), "") # ignore all sub-arguments after the first one
                    if template_suffix and set("bce").issuperset(template_suffix):
                        stem = f"{subopt}-{''.join(sorted(template_suffix))}"
                    else:
                        stem = subopt
                    template = read_template(stem, official_template_dir)
                
                # If the subopt was a relation template, use it to convert the source
                if template:
                    if template["extension"] == "sql":
                        if not sql_relations:
                            sql_source = source
                            sql_source = op_tk.run(sql_source, "ascii", {"labels": 1, "roles": 1}, self.params)
                            sql_mcd = Mcd(sql_source, self.get_font_metrics, **self.params)
                            sql_relations = Relations(sql_mcd, self.params)
                        text = sql_relations.get_text(template)
                    else:
                        if not raw_relations:
                            mcd = Mcd(source, self.get_font_metrics, **self.params)
                            raw_relations = Relations(mcd, self.params)
                        text = raw_relations.get_text(template)
                    result = {
                        "stem_suffix": template["stem_suffix"],
                        "text": text,
                        "extension": template["extension"],
                        "to_defer": template.get("to_defer", False),
                        "highlight": template.get("highlight", "plain"),
                    }

                # Fall back to a dynamically loaded conversion operation
                else:
                    try:
                        module = importlib.import_module(f".convert._{subopt}", package="mocodo")
                    except ModuleNotFoundError:
                        raise subopt_error("convert", subopt)
                    result = module.run(source, subargs, self.common)
                
                result["text_path"] = Path(f"{self.params['output_name']}{result['stem_suffix']}.{result['extension']}")
                result["text_path"].write_text(f"{result['text'].rstrip()}\n", encoding="utf8")
                self.printer.write(self.common.output_success_message(result["text_path"]))
                results.append(result)
            for result in results:
                converted_file_paths.extend(self.get_converted_file_paths(result))
            response.may_log("converted_file_paths", converted_file_paths)

        if not raw_relations: # if the MCD has not be calculated during the conversions
            mcd = Mcd(source, self.get_font_metrics, **self.params)
        self.control_for_overlaps(mcd)
        resulting_paths = dump_mcd_to_svg(mcd, self.common)  # potential side-effect: update *_geo.json
        for path in resulting_paths:
            self.printer.write(self.common.output_success_message(path))

    def get_rendering_service(self, extension):
        path = Path(self.params["script_directory"], "resources", "rendering_services.json")
        try:
            rendering_services = json.loads(path.read_text(encoding="utf8"))
        except FileNotFoundError:
            raise MocodoError(46, _('The file "{path}" is missing.').format(path=path))  # fmt: skip
        except json.decoder.JSONDecodeError:
            raise MocodoError(47, _('The file "{path}" is not a valid JSON file.').format(path=path))  # fmt: skip
        except Exception as err:
            raise MocodoError(48, _('The file "{path}" could not be read:\n{err}').format(path=path, err=err))  # fmt: skip
        try:
            return rendering_services[extension]
        except KeyError:
            raise MocodoError(49, _('No third-party rendering service for extension "{extension}". You may want to add one in "{path}".').format(extension=extension, path=path))

    def get_converted_file_paths(self, result):
        if result.get("to_defer") and "defer" in self.params:
            # This text must be rendered by a third-party service
            urllib = importlib.import_module("urllib")
            requests = importlib.import_module("requests")
            mimetypes = importlib.import_module("mimetypes")
            service = self.get_rendering_service(result["extension"])
            data = result["text"]
            for preprocessing in service.get("preprocessing", []):
                if preprocessing == "minify_graphviz":
                    # Spare some bandwidth
                    data = minify_graphviz(data)
                elif preprocessing == "urlsafe_encoding":
                    data = urlsafe_encoding(data)
                elif preprocessing == "encode_prefix":
                    # Sole use case (so far): "https://mocodo.net/?mcd="" becomes "https%3A//mocodo.net/%3Fmcd%3D"
                    index = data.find("=") + 1
                    data = urllib.parse.quote(data[:index]) + data[index:]
            url = service["url"].format(data=data)
            response = requests.get(url)
            if not response.ok:
                raise MocodoError(23, _("The HTTP status code {code} was returned by:\n{url}").format(code=response.status_code, url=url)) # fmt: skip
            content_type = response.headers['content-type']
            extension = mimetypes.guess_extension(content_type)
            resp_path = result["text_path"].with_suffix(extension)
            if content_type.startswith("text/"):
                resp_path.write_text(response.text, encoding="utf8")
            else:
                resp_path.write_bytes(response.content)
            yield str(resp_path)
        elif result.get("stem_suffix") == "_mld" and result.get("extension") == "mcd" and self.params["is_magic"]: # relational diagram
            mld = Mcd(result["text"], self.get_font_metrics, **self.params)
            backup_input = self.params["input"]
            backup_output_name = self.params["output_name"]
            self.params["input"] = result["text_path"]
            self.params["output_name"] = f"{self.params['output_name']}_mld"
            dump_mcd_to_svg(mld, self.common)
            self.params["input"] = backup_input
            self.params["output_name"] = backup_output_name
            yield str(result['text_path'].with_suffix(".svg"))
        else:
            # This text doesn't need further processing
            yield str(result["text_path"])

    def control_for_overlaps(self, mcd):
        if self.params["detect_overlaps"]:
            overlaps = mcd.get_overlaps()
            if overlaps:
                acc = []
                for (b1, b2, b3, b4) in overlaps:
                    if b3 == b4:
                        acc.append(_('  - Leg "{b1} — {b2}" overlaps "{b3}".').format(b1=b1, b2=b2, b3=b3))  # fmt: skip
                    else:
                        acc.append(_('  - Legs "{b1} — {b2}" and "{b3} — {b4}" overlap.').format(b1=b1, b2=b2, b3=b3, b4=b4))  # fmt: skip
                details = "\n".join(sorted(acc))
                raise MocodoError(29, _('Bad layout of boxes:\n{details}\nTo fix the problem, reorder and/or skip lines in the source text, either manually, or with the option -t arrange (chocolate bar under Mocodo online).').format(details=details))  # fmt: skip

    def add_gutter_params(self, params):
        gutters = dict(params.get("gutters", []))

        # Ensure that the --gutter sub-options are a subset of {"ids", "types"}
        for subopt in gutters:
            if subopt not in ("ids", "types"):
                raise subopt_error("gutters", subopt)
        
        # Create the sub-option "ids" if needed, and initialize id_gutter params
        if "ids" not in gutters:
            gutters["ids"] = {} # created at the end of the (ordered) dict
        params["id_gutter_visibility"] = gutters["ids"].get("visibility", "auto")
        params["id_gutter_weak_string"] = gutters["ids"].get("weak", "id")
        params["id_gutter_strong_string"] = gutters["ids"].get("strong", "ID")
        s = gutters["ids"].get("alts", "123456789")
        params["id_gutter_alts"] = dict(zip("123456789", s + "123456789"[len(s):]))

        # Create the sub-option "types" if needed, and initialize type_gutter params
        # NB: the gutter for types is not implemented yet, but all needed params are already there
        if "types" not in gutters:
            gutters["types"] = {}
        params["type_gutter_visibility"] = gutters["types"].get("visibility", "auto")

        # We are sure that gutters is a dict with keys "ids" and "types" only. Order matters.
        (params["left_gutter"], params["right_gutter"]) = tuple(gutters.keys())


def main():
    printer = Printer(accumulate=False)
    run = Runner(sys.argv[1:], printer)
    try:
        run()
    except MocodoError as err:
        print(str(err), file=sys.stderr)
        sys.exit(err.errno)


if __name__ == "__main__": # to be run with `python -m mocodo`
    main()



import textwrap


class MocodoError(Exception):
    def __init__(self, errno, message):
        message = f"Mocodo Err.{errno} - {message}"
        message = "\n".join(
            [
                "\n".join(
                    textwrap.wrap(
                        line,
                        88,
                        break_long_words=False,
                        replace_whitespace=False,
                    )
                )
                for line in message.splitlines()
                if line.strip() != ""
            ]
        )
        self.errno = errno
        super(MocodoError, self).__init__(message)

def subarg_error(subsubopt, subsubarg):
    return MocodoError(19, _('Invalid sub-argument: "{subsubopt}={subsubarg}".').format(subsubopt=subsubopt, subsubarg=subsubarg))  # fmt: skip

def subsubopt_error(subsubopt):
    return MocodoError(20, _('Invalid sub-sub-option: "{subsubopt}".').format(subsubopt=subsubopt))  # fmt: skip

def subopt_error(opt, subopt):
    raise MocodoError(21, _('Unknown "{opt}" sub-option: "{subopt}".').format(opt=opt, subopt=subopt))  # fmt: skip
